"use strict";
exports.__esModule = true;
var utils_1 = require("./utils");
// WARNING: When changing this number, you also have to adjust the easeDelta calculation below
exports.PERFECT_SCORE = 3;
exports.WARP_INTERVAL_IN_MONTHS = 3;
exports.initialSRSData = function (params) {
    if (params === void 0) { params = {}; }
    return ({
        lastAnsweredAt: null,
        dueAt: null,
        ease: params.randomizeInitiaEase ? 2.5 + (Math.random() * 0.5 - 0.25) : 2.5,
        intervalInMins: null,
        reps: 0,
        lapses: 0,
        repsSinceLastLapse: 0
    });
};
exports.answer = function (srsData, score, answeredAt) {
    if (score < 0) {
        throw new TypeError("score must not be negative; srsData: " + JSON.stringify(srsData) + "; score: " + score);
    }
    var lastAnsweredAt = srsData.lastAnsweredAt, dueAt = srsData.dueAt, ease = srsData.ease, intervalInMins = srsData.intervalInMins, reps = srsData.reps, lapses = srsData.lapses, repsSinceLastLapse = srsData.repsSinceLastLapse;
    var readinessFactor = 1;
    if (dueAt && lastAnsweredAt) {
        var scheduledInterval = +dueAt - +lastAnsweredAt;
        var actualInterval = +answeredAt - +lastAnsweredAt;
        readinessFactor = actualInterval / scheduledInterval;
    }
    if (readinessFactor < 0) {
        throw new TypeError("readinessFactor must not be negative; srsData: " + JSON.stringify(srsData) + "; score: " + score);
    }
    // Update ease
    var easeDelta = (-0.2 + 0.1 * score) * readinessFactor;
    var updatedEase = utils_1.clamp(ease + easeDelta, 1.3, 3.5);
    // Update interval
    var updatedInterval = Math.ceil(updateInterval(score, updatedEase, readinessFactor, reps, repsSinceLastLapse, intervalInMins));
    // Update rep counts
    var updatedReps = reps + 1;
    var updatedLapses, updatedRepsSinceLastLapse;
    if (score < Number.EPSILON) {
        updatedLapses = lapses + 1;
        updatedRepsSinceLastLapse = 0;
    }
    else {
        updatedLapses = lapses;
        updatedRepsSinceLastLapse = repsSinceLastLapse + 1;
    }
    // Calculate the next due at date
    var updatedDueAt = new Date(answeredAt.getTime() + updatedInterval * 60 * 1000);
    return {
        lastAnsweredAt: answeredAt,
        dueAt: updatedDueAt,
        ease: updatedEase,
        intervalInMins: updatedInterval,
        reps: updatedReps,
        lapses: updatedLapses,
        repsSinceLastLapse: updatedRepsSinceLastLapse
    };
};
var updateInterval = function (score, updatedEase, readinessFactor, reps, repsSinceLastLapse, intervalInMins) {
    if (score < Number.EPSILON) {
        return utils_1.daysInMins(1);
    }
    if (reps === 0) {
        if (score <= exports.PERFECT_SCORE / 3 + Number.EPSILON) {
            return utils_1.daysInMins(3);
        }
        else if (score <= (exports.PERFECT_SCORE * 2) / 3 + Number.EPSILON) {
            return utils_1.daysInMins(7);
        }
        return utils_1.daysInMins(14);
    }
    if (intervalInMins === null) {
        throw new TypeError("intervalInMins is null when it shouldn't be");
    }
    // TODO(jan): We can put this back in if we remove the randomness, e.g.
    // by multiplying the rep count with a number derived from the _id (which
    // would have to be a separate parameter to this function)
    // This adds +-5% randomness to the interval delta
    // const fuzzinessFactor = 1 + Math.random() * 0.1 - 0.05
    var intervalDelta = intervalInMins * (updatedEase - 1);
    // intervalDelta *= fuzzinessFactor
    intervalDelta *= readinessFactor;
    return Math.ceil(intervalInMins + intervalDelta);
};
exports.warp = function (srsData, answeredAt) { return ({
    lastAnsweredAt: answeredAt,
    dueAt: new Date(answeredAt.getTime() + utils_1.monthsInMins(exports.WARP_INTERVAL_IN_MONTHS) * 60 * 1000),
    ease: 3.5,
    intervalInMins: utils_1.monthsInMins(exports.WARP_INTERVAL_IN_MONTHS),
    reps: srsData.reps + 1,
    lapses: srsData.lapses,
    repsSinceLastLapse: srsData.repsSinceLastLapse + 1
}); };
exports.getPercentageMature = function (srsDataArray) {
    var intervals = srsDataArray.map(function (srsData) { return srsData.intervalInMins || 0; });
    var sum = intervals.reduce(function (a, b) { return a + b; }, 0);
    var avg = sum / intervals.length;
    return avg / utils_1.daysInMins(3 * 7);
};
